:root{--litestar-blue:#202235;--litestar-gold:#edb641;--litestar-yellow:#ffd480;--litestar-gray:#dcdfe4;--litestar-light:#fff;--litestar-dark:#202020;--accent-9:var(--litestar-gold);--syntax-light-property:var(--litestar-gold);--syntax-light-keyword:var(--litestar-blue);--syntax-light-constant:var(--litestar-blue);--syntax-light-text:var(--litestar-gold);--syntax-dark-property:var(--litestar-gold);--syntax-dark-keyword:var(--litestar-gray);--syntax-dark-constant:var(--litestar-gold);--syntax-dark-text:var(--litestar-gray)}#badges img{margin-top:0;margin-bottom:0}#home-head{display:flex;align-items:center;justify-content:space-between;-moz-column-gap:20px;column-gap:20px}@media (max-width:600px){#home-head{flex-wrap:wrap;justify-content:center}#home-head a img{width:250px;height:250px}}#home-head h1{font-size:4.2rem;margin-top:3.4rem;margin-bottom:1rem;font-weight:700;line-height:1.2}#badges{display:flex;flex-wrap:wrap;gap:10px;margin-bottom:3em}.sponsor{border-radius:10px;width:auto;max-height:150px;display:flex;flex-direction:column;align-items:center;text-align:center}
